# ------------------------------------------------------------------------------
# Unit tests for the clorm ORM interface
# ------------------------------------------------------------------------------

import importlib
import os
import sys
import unittest

import clingo

from clorm.orm import noclingo
from clorm.orm.noclingo import (
    Function,
    Number,
    String,
    SymbolMode,
    clingo_to_noclingo,
    get_Infimum,
    get_Supremum,
    get_symbol_mode,
    noclingo_to_clingo,
    set_symbol_mode,
)

from .support import check_errmsg

clingo_version = clingo.__version__

# Error messages for CPython and PyPy vary
PYPY = sys.implementation.name == "pypy"

# ------------------------------------------------------------------------------
# ------------------------------------------------------------------------------

__all__ = ["NoClingoTestCase", "NoClingoDisabledTestCase"]

# ------------------------------------------------------------------------------
# Tests that don't require that NOCLINGO is enabled
# ------------------------------------------------------------------------------


class NoClingoTestCase(unittest.TestCase):
    def setUp(self):
        pass

    def tearDown(self):
        pass

    # --------------------------------------------------------------------------
    #
    # --------------------------------------------------------------------------
    def test_invalid_values(self):

        # The symbol type needs to be valid
        with self.assertRaises(TypeError) as ctx:
            noclingo.NoSymbol(6, 4)

    # --------------------------------------------------------------------------
    #
    # --------------------------------------------------------------------------
    def test_string(self):
        nc = noclingo.NoString("atest")
        c = clingo.String("atest")
        self.assertEqual(nc.string, c.string)
        self.assertEqual(str(nc), str(c))
        self.assertEqual(nc.type, noclingo.SymbolType.String)

    def test_number(self):
        nc = noclingo.NoNumber(1)
        c = clingo.Number(1)
        self.assertEqual(nc.number, c.number)
        self.assertEqual(str(nc), str(c))
        self.assertEqual(nc.type, noclingo.SymbolType.Number)

    def test_infimum_supremum(self):
        nc1 = noclingo.NoInfimum
        nc2 = noclingo.NoNumber(1)
        nc3 = noclingo.NoString("a")
        nc4 = noclingo.NoFunction("a")
        nc5 = noclingo.NoSupremum

        self.assertTrue(nc1 == nc1)
        self.assertTrue(nc1 <= nc1)
        self.assertTrue(nc1 < nc2)
        self.assertTrue(nc1 < nc3)
        self.assertTrue(nc1 < nc4)
        self.assertTrue(nc1 < nc5)
        self.assertTrue(nc5 > nc1)
        self.assertTrue(nc5 > nc2)
        self.assertTrue(nc5 > nc3)
        self.assertTrue(nc5 > nc4)
        self.assertTrue(nc5 >= nc5)
        self.assertTrue(nc1 != nc2)

        self.assertEqual(str(nc1), str(clingo.Infimum))
        self.assertEqual(str(nc5), str(clingo.Supremum))
        self.assertEqual(nc1.type, noclingo.SymbolType.Infimum)
        self.assertEqual(nc5.type, noclingo.SymbolType.Supremum)

    def test_function(self):
        nc = noclingo.NoFunction("atest")
        c = clingo.Function("atest")
        self.assertEqual(str(nc), str(c))
        self.assertEqual(nc.type, noclingo.SymbolType.Function)

        nc1 = noclingo.NoFunction("aaaa")
        nc2 = noclingo.NoString("bbb")
        nc3 = noclingo.NoFunction("ccc", [noclingo.NoNumber(10)])
        c1 = clingo.Function("aaaa")
        c2 = clingo.String("bbb")
        c3 = clingo.Function("ccc", [clingo.Number(10)])

        nc = noclingo.NoFunction("atest", [nc1, nc2, nc3])
        c = clingo.Function("atest", [c1, c2, c3])

        self.assertEqual(str(nc), str(c))
        self.assertEqual(nc.name, c.name)
        self.assertEqual(nc.name, "atest")
        self.assertEqual(len(nc.arguments), len(c.arguments))
        self.assertEqual(nc.arguments[0].name, c.arguments[0].name)
        self.assertEqual(nc.arguments[0].name, "aaaa")
        self.assertEqual(nc.arguments[1].string, c.arguments[1].string)

        nc4 = noclingo.NoFunction("ccc", [noclingo.NoNumber(10)], False)
        c4 = clingo.Function("ccc", [clingo.Number(10)], False)
        self.assertEqual(str(nc4), str(c4))
        self.assertEqual(nc4.positive, c4.positive)
        self.assertEqual(nc4.negative, c4.negative)

    def test_tuple(self):
        c_const = clingo.Function("aaaa")
        nc_const = noclingo.NoFunction("aaaa")
        c_string = clingo.String("bbb")
        nc_string = noclingo.NoString("bbb")

        # Check using the convenience Tuple_() function
        self.assertEqual(noclingo.NoTuple_([]), noclingo.NoFunction("", []))
        self.assertEqual(noclingo.NoTuple_([c_const]), noclingo.NoFunction("", [c_const]))

        # clingo.NoTuple_() has no default so noclingo.NoTuple_() acts the same
        # self.assertEqual(clingo.NoTuple_(), clingo.NoTuple_([]))
        with self.assertRaises(TypeError) as ctx:
            self.assertEqual(noclingo.NoTuple_(), noclingo.NoTuple_([]))

        check_errmsg("NoTuple_() missing 1 required positional argument: 'arguments'", ctx)

        c_empty_tuple = clingo.Tuple_([])
        nc_empty_tuple = noclingo.NoTuple_([])
        c_one_tuple = clingo.Tuple_([c_const])
        nc_one_tuple = noclingo.NoTuple_([nc_const])

        # Check special cases of empty and singleton tuples
        self.assertEqual(str(c_empty_tuple), str(nc_empty_tuple))
        self.assertEqual(str(nc_one_tuple), str(c_one_tuple))

        # tuple with two or more elements
        self.assertEqual(
            str(clingo.Tuple_([c_const, c_string])), str(noclingo.NoTuple_([nc_const, nc_string]))
        )

    def test_hash_and_equality_comparison_ops(self):
        nc1 = noclingo.NoString("aaaGGDFa")
        nc2 = noclingo.NoString("aaaGGDFa")
        self.assertEqual(hash(nc1), hash(nc2))
        self.assertTrue(nc1 == nc2)
        self.assertTrue(nc1 <= nc2)
        self.assertTrue(nc1 >= nc2)

        nc1 = noclingo.NoNumber(34068390)
        nc2 = noclingo.NoNumber(34068390)
        self.assertEqual(hash(nc1), hash(nc2))
        self.assertTrue(nc1 == nc2)
        self.assertTrue(nc1 <= nc2)
        self.assertTrue(nc1 >= nc2)

        nc1 = noclingo.NoFunction("aaaa")
        nc2 = noclingo.NoFunction("aaaa")
        self.assertEqual(hash(nc1), hash(nc2))
        self.assertTrue(nc1 == nc2)
        self.assertTrue(nc1 <= nc2)
        self.assertTrue(nc1 >= nc2)

        nc1 = noclingo.NoFunction("aaaa", [], False)
        nc2 = noclingo.NoFunction("aaaa", [], False)
        self.assertEqual(hash(nc1), hash(nc2))
        self.assertTrue(nc1 == nc2)
        self.assertTrue(nc1 <= nc2)
        self.assertTrue(nc1 >= nc2)

        a1 = noclingo.NoFunction("aaaa")
        a2 = noclingo.NoFunction("aaaa")
        b1 = noclingo.NoString("bbb")
        b2 = noclingo.NoString("bbb")
        c1 = noclingo.NoNumber(45)
        c2 = noclingo.NoNumber(45)
        d1 = noclingo.NoFunction("dfdf", [c1])
        d2 = noclingo.NoFunction("dfdf", [c2])
        nc1 = noclingo.NoFunction("ccc", [a1, b1, c1, d1])
        nc2 = noclingo.NoFunction("ccc", [a2, b2, c2, d2])
        self.assertEqual(hash(nc1), hash(nc2))
        self.assertTrue(nc1 == nc2)
        self.assertTrue(nc1 <= nc2)
        self.assertTrue(nc1 >= nc2)

    def test_comparison_ops(self):
        nc1 = noclingo.NoNumber(34)
        nc2 = noclingo.NoNumber(43)
        self.assertTrue(nc1 <= nc2)
        self.assertTrue(nc1 < nc2)
        self.assertTrue(nc2 >= nc1)
        self.assertTrue(nc2 > nc1)

        nc3 = noclingo.NoString("abcd")
        nc4 = noclingo.NoString("bcde")
        self.assertTrue(nc3 <= nc4)
        self.assertTrue(nc3 < nc4)
        self.assertTrue(nc4 >= nc3)
        self.assertTrue(nc4 > nc3)

        nc5 = noclingo.NoFunction("abc", [noclingo.NoNumber(45)])
        nc6 = noclingo.NoFunction("abc", [noclingo.NoString("45")])
        c5 = clingo.Function("abc", [clingo.Number(45)])
        c6 = clingo.Function("abc", [clingo.String("45")])
        if c5 < c6:
            self.assertTrue(nc5 < nc6)
        else:
            self.assertTrue(nc5 > nc6)

        nc7 = noclingo.NoFunction("abc", [noclingo.NoString("45"), noclingo.NoNumber(5)])
        self.assertTrue(nc6 < nc7)

        def compare_ordering(a, b):
            if noclingo_to_clingo(a) < noclingo_to_clingo(b):
                self.assertTrue(a < b)
            elif noclingo_to_clingo(b) < noclingo_to_clingo(a):
                self.assertTrue(b < a)
            else:
                self.assertEqual(a, b)

        compare_ordering(noclingo.NoString("1"), noclingo.NoNumber(2))
        compare_ordering(noclingo.NoNumber(1), noclingo.NoString("2"))
        compare_ordering(noclingo.NoString("1"), noclingo.NoFunction("2"))
        compare_ordering(noclingo.NoFunction("2"), noclingo.NoString("1"))
        compare_ordering(noclingo.NoNumber(1), noclingo.NoFunction("2"))
        compare_ordering(noclingo.NoFunction("1"), noclingo.NoNumber(2))
        compare_ordering(noclingo.NoInfimum, noclingo.NoSupremum)
        compare_ordering(noclingo.NoSupremum, noclingo.NoInfimum)
        compare_ordering(noclingo.NoInfimum, noclingo.NoNumber(2))
        compare_ordering(noclingo.NoInfimum, noclingo.NoString("2"))
        compare_ordering(noclingo.NoInfimum, noclingo.NoFunction("2"))
        compare_ordering(noclingo.NoSupremum, noclingo.NoFunction("2"))
        compare_ordering(noclingo.NoSupremum, noclingo.NoString("2"))
        compare_ordering(noclingo.NoSupremum, noclingo.NoNumber(2))

    def test_same_error_messages(self):
        def _get_error(func, *args):
            try:
                func(*args)
            except (TypeError, AttributeError, ValueError) as e:
                return e

        def _assertEqualError(c_error, nc_error):
            self.assertEqual(type(c_error), type(nc_error))
            # NOTE: Not testing the exact error message because PyPy and CPython produce
            # different error messages and I don't know how to use python itself to generate
            # the message.
            #            self.assertEqual(str(c_error), str(nc_error))

        _assertEqualError(_get_error(clingo.Number, "a"), _get_error(noclingo.NoNumber, "a"))
        _assertEqualError(_get_error(clingo.Number, ["a"]), _get_error(noclingo.NoNumber, ["a"]))
        _assertEqualError(_get_error(clingo.String, 1), _get_error(noclingo.String, 1))
        _assertEqualError(_get_error(clingo.String, [1]), _get_error(noclingo.String, [1]))

    # NOTE: I think in clingo 5.5 the comparison operators correctly return
    # NotImplemented so that we can implement the comparison between Symbol and
    # NoSymbol objects.
    def _clingo_pre5_5_test_clingo_noclingo_difference(self):
        self.assertNotEqual(clingo.String("blah"), noclingo.NoString("blah"))
        self.assertNotEqual(clingo.Number(5), noclingo.NoNumber(5))

    def test_clingo_to_noclingo(self):

        # Converting the Infimum and Supremum
        cli = clingo.Infimum
        ncli = noclingo.NoInfimum
        cls = clingo.Supremum
        ncls = noclingo.NoSupremum

        self.assertEqual(clingo_to_noclingo(cli), ncli)
        self.assertEqual(clingo_to_noclingo(cls), ncls)
        self.assertEqual(noclingo_to_clingo(ncli), cli)
        self.assertEqual(noclingo_to_clingo(ncls), cls)

        # Converting simple structures
        cl1 = clingo.Function("const")
        ncl1 = noclingo.NoFunction("const")
        cl2 = clingo.Number(3)
        ncl2 = noclingo.NoNumber(3)
        cl3 = clingo.String("No")
        ncl3 = noclingo.NoString("No")

        self.assertEqual(clingo_to_noclingo(cl1), ncl1)
        self.assertEqual(clingo_to_noclingo(cl2), ncl2)
        self.assertEqual(clingo_to_noclingo(cl3), ncl3)
        self.assertEqual(noclingo_to_clingo(ncl1), cl1)
        self.assertEqual(noclingo_to_clingo(ncl2), cl2)
        self.assertEqual(noclingo_to_clingo(ncl3), cl3)

        # More complex function structures
        cl4 = clingo.Function("", [cl1, cl2])
        ncl4 = noclingo.NoFunction("", [ncl1, ncl2])
        self.assertEqual(clingo_to_noclingo(cl4), ncl4)
        self.assertEqual(noclingo_to_clingo(ncl4), cl4)

        cl5 = clingo.Function("f", [cl3, cl4, cl1], False)
        ncl5 = noclingo.NoFunction("f", [ncl3, ncl4, ncl1], False)
        self.assertEqual(clingo_to_noclingo(cl5), ncl5)
        self.assertEqual(noclingo_to_clingo(ncl5), cl5)

        # If it is already the correct symbol type then no conversion required
        self.assertEqual(clingo_to_noclingo(ncl1), ncl1)
        self.assertEqual(noclingo_to_clingo(cl1), cl1)
        self.assertTrue(clingo_to_noclingo(ncl1) is ncl1)
        self.assertTrue(noclingo_to_clingo(cl1) is cl1)

    def test_comparison_nosymbol_and_symbol(self):
        cl_infimum = clingo.Infimum
        ncl_infimum = noclingo.NoInfimum
        cl_supremum = clingo.Supremum
        ncl_supremum = noclingo.NoSupremum
        cl_constant1 = clingo.Function("const1")
        ncl_constant1 = noclingo.NoFunction("const1")
        cl_constant2 = clingo.Function("const2")
        ncl_constant2 = noclingo.NoFunction("const2")
        cl_number1 = clingo.Number(1)
        ncl_number1 = noclingo.NoNumber(1)
        cl_number2 = clingo.Number(2)
        ncl_number2 = noclingo.NoNumber(2)
        cl_string1 = clingo.String("No1")
        ncl_string1 = noclingo.NoString("No1")
        cl_string1 = clingo.String("No2")
        ncl_string1 = noclingo.NoString("No2")

        self.assertTrue(ncl_infimum == cl_infimum)
        self.assertTrue(cl_infimum == ncl_infimum)
        self.assertTrue(ncl_constant1 == cl_constant1)
        self.assertTrue(cl_constant1 == ncl_constant1)
        self.assertTrue(ncl_constant1 == cl_constant1)

        c_x = clingo.Function("b", positive=False)
        c_y = clingo.Function("a", positive=True)
        nc_x = noclingo.NoFunction("b", positive=False)
        nc_y = noclingo.NoFunction("a", positive=True)

        self.assertTrue(c_x > c_y)
        self.assertTrue(nc_x > nc_y)
        self.assertTrue(c_x > nc_y)

    def test_symbol_modes(self):
        # By default CLINGO mode
        #        self.assertEqual(get_symbol_mode(), SymbolMode.CLINGO)

        cli = clingo.Infimum
        cls = clingo.Supremum
        cl1 = clingo.Function("const")
        cl2 = clingo.Number(3)
        cl3 = clingo.String("No")
        cl4 = clingo.Function("", [cl1, cl2])
        cl5 = clingo.Function("f", [cl3, cl4, cl1], False)

        ncli = noclingo.NoInfimum
        ncls = noclingo.NoSupremum
        ncl1 = noclingo.NoFunction("const")
        ncl2 = noclingo.NoNumber(3)
        ncl3 = noclingo.NoString("No")
        ncl4 = noclingo.NoFunction("", [ncl1, ncl2])
        ncl5 = noclingo.NoFunction("f", [ncl3, ncl4, ncl1], False)

        set_symbol_mode(SymbolMode.CLINGO)
        self.assertEqual(get_symbol_mode(), SymbolMode.CLINGO)

        expect_cli = get_Infimum()
        expect_cls = get_Supremum()
        expect_cl1 = Function("const")
        expect_cl2 = Number(3)
        expect_cl3 = String("No")
        expect_cl4 = Function("", [expect_cl1, expect_cl2])
        expect_cl5 = Function("f", [expect_cl3, expect_cl4, expect_cl1], False)

        set_symbol_mode(SymbolMode.NOCLINGO)
        self.assertEqual(get_symbol_mode(), SymbolMode.NOCLINGO)

        expect_ncli = get_Infimum()
        expect_ncls = get_Supremum()
        expect_ncl1 = Function("const")
        expect_ncl2 = Number(3)
        expect_ncl3 = String("No")
        expect_ncl4 = Function("", [expect_ncl1, expect_ncl2])
        expect_ncl5 = Function("f", [expect_ncl3, expect_ncl4, expect_ncl1], False)

        # Set back to the default
        set_symbol_mode(SymbolMode.CLINGO)

        self.assertEqual(cli, expect_cli)
        self.assertEqual(cls, expect_cls)
        self.assertEqual(cl1, expect_cl1)
        self.assertEqual(cl2, expect_cl2)
        self.assertEqual(cl3, expect_cl3)
        self.assertEqual(cl4, expect_cl4)
        self.assertEqual(cl5, expect_cl5)

        self.assertEqual(ncli, expect_ncli)
        self.assertEqual(ncls, expect_ncls)
        self.assertEqual(ncl1, expect_ncl1)
        self.assertEqual(ncl2, expect_ncl2)
        self.assertEqual(ncl3, expect_ncl3)
        self.assertEqual(ncl4, expect_ncl4)
        self.assertEqual(ncl5, expect_ncl5)


# ------------------------------------------------------------------------------
# Tests that require CLORM_NOCLINGO to be disabled. When CLORM_NOCLINGO is
# disabled then set_symbol_mode() raises an error. And noclingo.Function(),
# noclingo.Tuple_(), noclingo.String(), noclingo.Number() all point directly to
# the clingo functions.
# ------------------------------------------------------------------------------


class NoClingoDisabledTestCase(unittest.TestCase):
    def setUp(self):
        pass

    def tearDown(self):
        pass

    def test_noclingo_switch_disabled(self):
        os.environ["CLORM_NOCLINGO"] = "False"
        import clorm.orm.noclingo

        importlib.reload(clorm.orm.noclingo)

        self.assertFalse(clorm.orm.noclingo.ENABLE_NOCLINGO)

        # set_symbol_mode will raise an error
        with self.assertRaises(RuntimeError) as ctx:
            clorm.orm.noclingo.set_symbol_mode(SymbolMode.CLINGO)
        check_errmsg("NOCLINGO mode is disabled.", ctx)

        # The various noclingo functions point to the real clingo functions
        self.assertEqual(clorm.orm.noclingo.Function, clingo.Function)
        self.assertEqual(clorm.orm.noclingo.Tuple_, clingo.Tuple_)
        self.assertEqual(clorm.orm.noclingo.String, clingo.String)
        self.assertEqual(clorm.orm.noclingo.Number, clingo.Number)

        os.environ["CLORM_NOCLINGO"] = "True"
        import clorm.orm.noclingo

        importlib.reload(clorm.orm.noclingo)


# ------------------------------------------------------------------------------
# main
# ------------------------------------------------------------------------------
if __name__ == "__main__":
    raise RuntimeError("Cannot run modules")



# ------------------------------------------------------------------------------
# Unit tests for Clorm ORM @-syntax integration (calling Python from ASP).
#
# Note: I'm trying to clearly separate tests of the official Clorm API from
# tests of the internal implementation. Tests for the API have names
# "test_api_XXX" while non-API tests are named "test_nonapi_XXX". This is still
# to be completed.
# ------------------------------------------------------------------------------

import calendar
import datetime
import sys
import unittest
from typing import Tuple

from clingo import Function, Number, String, Tuple_
from clingo import __version__ as clingo_version

# Official Clorm API imports
# Official Clorm API imports
from clorm.orm import (
    ComplexTerm,
    ConstantField,
    ContextBuilder,
    IntegerField,
    StringField,
    TypeCastSignature,
    make_function_asp_callable,
    make_method_asp_callable,
)

# Implementation imports
from clorm.orm.atsyntax import _get_annotations

from .support import check_errmsg

# Error messages for CPython and PyPy vary
PYPY = sys.implementation.name == "pypy"

# ------------------------------------------------------------------------------
# ------------------------------------------------------------------------------

__all__ = [
    "TypeCastSignatureTestCase",
    "ContextBuilderTestCase",
]

# ------------------------------------------------------------------------------
#
# ------------------------------------------------------------------------------


class TypeCastSignatureTestCase(unittest.TestCase):
    def setUp(self):
        class DateField(StringField):
            pytocl = lambda dt: datetime.datetime.strftime(dt, "%Y%m%d")
            cltopy = lambda s: datetime.datetime.strptime(s, "%Y%m%d").date()

        class DowField(ConstantField):
            pytocl = lambda dt: calendar.day_name[dt.weekday()].lower()

        class EDate(ComplexTerm):
            idx = IntegerField()
            date = DateField()

            class Meta:
                name = "edate"

        self.DateField = DateField
        self.DowField = DowField
        self.EDate = EDate

    # --------------------------------------------------------------------------
    # Test the signature generation for writing python functions that can be
    # called from ASP.
    # --------------------------------------------------------------------------

    def test_signature(self):

        DateField = self.DateField
        DowField = self.DowField
        EDate = self.EDate

        sig1 = TypeCastSignature(DateField)  # returns a single date
        sig2 = TypeCastSignature([DateField])  # returns a list of dates
        sig3 = TypeCastSignature(DateField, DowField)  # takes a date and returns the day or week

        sig4 = TypeCastSignature(EDate.Field, EDate.Field)  # takes an EDate and returns an EDate

        with self.assertRaises(TypeError) as ctx:
            sig5 = TypeCastSignature(DateField, [int])
        with self.assertRaises(TypeError) as ctx:
            sig5 = TypeCastSignature(DateField, [DateField, DateField])

        date1 = datetime.date(2018, 1, 1)
        date2 = datetime.date(2019, 2, 2)

        edate1 = EDate(idx=1, date=date1)
        edate2 = EDate(idx=2, date=date2)

        # Test simple output and list output

        def getdate1():
            return date1

        def getdates():
            return [date1, date2]

        cl_getdate1 = sig1.wrap_function(getdate1)
        cl_getdates = sig2.wrap_function(getdates)
        self.assertEqual(cl_getdate1(), String("20180101"))
        self.assertEqual(cl_getdates(), [String("20180101"), String("20190202")])

        # Use decoractor mode

        @sig3.wrap_function
        def getdow(dt):
            return dt

        result = getdow(String("20180101"))
        self.assertEqual(result, Function("monday", []))

        # Test a ComplexTerm input and output
        @sig4.wrap_function
        def getedate(indate):
            return indate

        self.assertEqual(getedate(edate1.raw), edate1.raw)
        self.assertEqual(getedate(edate2.raw), edate2.raw)

        # Now test the method wrapper
        class Tmp(object):
            def __init__(self, x, y):
                self._x = x
                self._y = y

            def get_pair(self):
                return [self._x, self._y]

            cl_get_pair = sig2.wrap_method(get_pair)

        t = Tmp(date1, date2)
        self.assertEqual(t.cl_get_pair(), [String("20180101"), String("20190202")])

    # --------------------------------------------------------------------------
    # Test the extended signatures with tuples
    # --------------------------------------------------------------------------

    def test_signature_with_tuples(self):
        DateField = self.DateField

        # Some complicated signatures
        sig1 = TypeCastSignature((IntegerField, DateField), (IntegerField, DateField))
        sig2 = TypeCastSignature(DateField, [(IntegerField, DateField)])

        @sig1.wrap_function
        def test_sig1(pair):
            return (pair[0], pair[1])

        @sig2.wrap_function
        def test_sig2(dt):
            return [(1, dt), (2, dt)]

        s_raw = String("20180101")
        t1_raw = Function("", [Number(1), s_raw])
        t2_raw = Function("", [Number(2), s_raw])

        #        result = test_sig1(t1_raw)
        self.assertEqual(test_sig1(t1_raw), t1_raw)
        self.assertEqual(test_sig2(s_raw), [t1_raw, t2_raw])

    # --------------------------------------------------------------------------
    # Test using function annotations and reporting better errors
    # --------------------------------------------------------------------------
    def test_get_annotations_errors(self):

        IF = IntegerField

        with self.assertRaises(TypeError) as ctx:

            def bad() -> IF:
                return 1

            s = _get_annotations(bad, True)
        check_errmsg("Cannot ignore", ctx)

        with self.assertRaises(TypeError) as ctx:

            def bad(a: IF, b: IF):
                return 1

            s = _get_annotations(bad)
        check_errmsg("Missing function", ctx)

        with self.assertRaises(TypeError) as ctx:

            def bad(a, b) -> IF:
                return 1

            s = _get_annotations(bad)
        check_errmsg("Missing type cast", ctx)

        with self.assertRaises(TypeError) as ctx:

            def bad(a: IF, b) -> IF:
                return 1

            s = _get_annotations(bad)
        check_errmsg("Missing type cast", ctx)

    # --------------------------------------------------------------------------
    # Test the signature generation for writing python functions that can be
    # called from ASP.
    # --------------------------------------------------------------------------

    def test_make_function_asp_callable(self):

        DateField = self.DateField
        DowField = self.DowField
        EDate = self.EDate

        date1 = datetime.date(2018, 1, 1)
        date2 = datetime.date(2019, 2, 2)

        edate1 = EDate(idx=1, date=date1)
        edate2 = EDate(idx=2, date=date2)

        def getdate1():
            return date1

        def getdates():
            return [date1, date2]

        # Test wrapper as a normal function and specifying a signature
        cl_getdate1 = make_function_asp_callable(DateField, getdate1)
        self.assertEqual(cl_getdate1(), String("20180101"))

        cl_getdates = make_function_asp_callable([DateField], getdates)
        self.assertEqual(cl_getdates(), [String("20180101"), String("20190202")])

        # Test wrapper as a decorator and specifying a signature
        @make_function_asp_callable(DateField)
        def getdate1():
            return date1

        self.assertEqual(getdate1(), String("20180101"))

        @make_function_asp_callable([DateField])
        def getdates():
            return [date1, date2]

        self.assertEqual(getdates(), [String("20180101"), String("20190202")])

        @make_function_asp_callable
        def getdates2(x: DateField, y: EDate.Field) -> [DateField]:
            """GETDATES2"""
            return [date1, date2]

        self.assertEqual(
            getdates2(String("20180101"), edate1.raw), [String("20180101"), String("20190202")]
        )
        self.assertEqual(getdates2.__doc__, """GETDATES2""")

        # Test wrapper as a decorator and use python type annotations
        @make_function_asp_callable
        def method(x: int, y: str) -> Tuple[int, int]:
            return (x, int(y))

        self.assertEqual(method(Number(42), String("24")), Function("", [Number(42), Number(24)]))

        with self.assertRaises(TypeError) as ctx:

            @make_function_asp_callable
            def getdates3(x, y):
                return [date1, date2]

        with self.assertRaises(TypeError) as ctx:

            @make_function_asp_callable
            def getdates4(x: DateField, y: DateField):
                return [date1, date2]

        # Now test the method wrapper
        class Tmp(object):
            def __init__(self, x, y):
                self._x = x
                self._y = y

            def get_pair(self):
                return [self._x, self._y]

            cl_get_pair = make_method_asp_callable([DateField], get_pair)

            @make_method_asp_callable
            def get_pair2(self) -> [DateField]:
                return [self._x, self._y]

        t = Tmp(date1, date2)
        self.assertEqual(t.cl_get_pair(), [String("20180101"), String("20190202")])
        self.assertEqual(t.get_pair2(), [String("20180101"), String("20190202")])

    def test_make_function_asp_callable_with_tuples(self):
        DateField = self.DateField

        # Some complicated signatures
        sig1 = TypeCastSignature((IntegerField, DateField), (IntegerField, DateField))
        sig2 = TypeCastSignature(DateField, [(IntegerField, DateField)])

        @make_function_asp_callable
        def test_sig1(pair: (IntegerField, DateField)) -> (IntegerField, DateField):
            return (pair[0], pair[1])

        @make_function_asp_callable
        def test_sig2(dt: DateField) -> [(IntegerField, DateField)]:
            return [(1, dt), (2, dt)]

        s_raw = String("20180101")
        t1_raw = Function("", [Number(1), s_raw])
        t2_raw = Function("", [Number(2), s_raw])

        #        result = test_sig1(t1_raw)
        self.assertEqual(test_sig1(t1_raw), t1_raw)
        self.assertEqual(test_sig2(s_raw), [t1_raw, t2_raw])

    def test_make_function_asp_callable_with_type_annotations(self):

        # Some complicated signatures

        @make_function_asp_callable
        def _sig1(x: int, y: int) -> int:
            return x + y

        @make_function_asp_callable
        def _sig2(pair: Tuple[int, int]) -> Tuple[int, int]:
            return (pair[1], pair[0])

        self.assertEqual(_sig1(Number(1), Number(2)), Number(3))
        self.assertEqual(_sig2(Tuple_([Number(1), Number(2)])), Tuple_([Number(2), Number(1)]))

    # --------------------------------------------------------------------------
    # Improving the error messages generated by the wrappers
    # --------------------------------------------------------------------------
    def test_make_function_asp_callable_error_feedback(self):
        @make_function_asp_callable
        def test_sig1(v: IntegerField) -> IntegerField:
            return float(v)

        @make_function_asp_callable
        def test_sig2(v: IntegerField) -> IntegerField:
            raise ValueError("Error")

        with self.assertRaises(TypeError) as ctx:
            test_sig1(Number(1), Number(2))
        check_errmsg("test_sig1() takes 1 positional arguments but 2 ", ctx)

        if clingo_version >= "5.5.0":
            with self.assertRaises(TypeError) as ctx:
                test_sig1(Number(1))
            if PYPY:
                check_errmsg("expected integer, got float object for output of test_sig1()", ctx)
            else:
                check_errmsg("an integer is required for output of test_sig1()", ctx)

        with self.assertRaises(ValueError) as ctx:
            test_sig2(Number(1))
        check_errmsg("Error: raised by test_sig2()", ctx)

    def test_make_method_asp_callable_error_feedback(self):
        class Tmp(object):
            def test1(self, v):
                return v

            @make_method_asp_callable
            def test_sig1(self, v: IntegerField) -> IntegerField:
                return float(v)

            @make_method_asp_callable
            def test_sig2(self, v: IntegerField) -> IntegerField:
                raise ValueError("Error")

        tmp = Tmp()

        with self.assertRaises(TypeError) as ctx:
            tmp.test_sig1(Number(1), Number(2))
        check_errmsg("test_sig1() takes 2 positional arguments but 3 ", ctx)

        if clingo_version >= "5.5.0":
            with self.assertRaises(TypeError) as ctx:
                tmp.test_sig1(Number(1))
            if PYPY:
                check_errmsg("expected integer, got float object for output of test_sig1()", ctx)
            else:
                check_errmsg("an integer is required for output of test_sig1()", ctx)

        with self.assertRaises(ValueError) as ctx:
            tmp.test_sig2(Number(1))
        check_errmsg("Error: raised by test_sig2()", ctx)

    # --------------------------------------------------------------------------
    # Test that the input signature can be hashed
    # --------------------------------------------------------------------------
    def test_input_signature(self):
        DateField = self.DateField

        # Some complicated signatures
        sig1 = TypeCastSignature((IntegerField, DateField), (IntegerField, DateField))
        sig2 = TypeCastSignature(DateField, [(IntegerField, DateField)])
        sigs = {}
        sigs[sig1.input_signature] = sig1
        sigs[sig2.input_signature] = sig2
        self.assertEqual(sigs[sig1.input_signature], sig1)
        self.assertEqual(sigs[sig2.input_signature], sig2)


# ------------------------------------------------------------------------------
# Tests for the ContextBuilder
# ------------------------------------------------------------------------------


class ContextBuilderTestCase(unittest.TestCase):
    def setUp(self):
        pass

    def test_register(self):

        SF = StringField
        IF = IntegerField
        CF = ConstantField
        # Functions to add to the context
        def add(a: IF, b: IF) -> IF:
            return a + b

        def mirror(val: CF) -> CF:
            return val

        self.assertEqual(add(1, 4), 5)
        self.assertEqual(mirror("aname"), "aname")

        # Test the register function as a non-decorator
        cb1 = ContextBuilder()
        cb1.register(add)
        cb1.register(mirror)
        ctx1 = cb1.make_context()
        self.assertEqual(type(ctx1).__name__, "Context")
        self.assertTrue("add" in type(ctx1).__dict__)
        self.assertTrue("mirror" in type(ctx1).__dict__)

        n1 = Number(1)
        n2 = Number(2)
        n3 = Number(3)
        n4 = Number(4)
        c1 = Function("aname")
        self.assertEqual(ctx1.add(n1, n3), n4)
        self.assertEqual(ctx1.mirror(c1), c1)

        # Test registering functions but using external signatures
        def add2(a, b):
            return a + b

        def mirror2(val):
            return val

        cb2 = ContextBuilder()
        cb2.register(IF, IF, IF, add2)
        cb2.register(CF, CF, mirror2)
        ctx2 = cb2.make_context("Ctx2")
        self.assertEqual(type(ctx2).__name__, "Ctx2")
        self.assertEqual(ctx2.add2(n1, n3), n4)
        self.assertEqual(ctx2.mirror2(c1), c1)

        # Test the register function as a decorator
        cb3 = ContextBuilder()

        @cb3.register
        def add2(a: IF, b: IF) -> IF:
            return a + b

        self.assertEqual(add2(1, 2), 3)

        @cb3.register(IF, IF, IF)
        def add4(a, b):
            return a + b

        self.assertEqual(add4(1, 2), 3)

        ctx3 = cb3.make_context()
        self.assertEqual(ctx3.add2(n1, n2), n3)
        self.assertEqual(ctx3.add4(n1, n2), n3)

        # Test the register function in decorator mode where the function
        # returns a list.
        cb4 = ContextBuilder()

        @cb4.register
        def arange1(start: IF, end: IF) -> [IF]:
            return list(range(start, end))

        self.assertEqual(arange1(1, 2), [1])

        @cb4.register(IF, IF, [IF])
        def arange2(start, end):
            return list(range(start, end))

        self.assertEqual(arange1(1, 2), [1])

        @cb4.register([IF])
        def fixedrange():
            return list(range(1, 2))

        self.assertEqual(fixedrange(), [1])

        # Must sure we can register to return a list of tuples (bug issue #42)
        @cb4.register([(IF, IF)])
        def fixedrange2():
            return [(1, 1)]

        self.assertEqual(fixedrange2(), [(1, 1)])

        @cb4.register_name("blah", [(IF, IF)])
        def fixedrange3():
            return [(1, 1)]

        self.assertEqual(fixedrange3(), [(1, 1)])

        ctx4 = cb4.make_context()
        self.assertEqual(ctx4.arange1(n1, n2), [n1])
        self.assertEqual(ctx4.arange2(n1, n2), [n1])
        self.assertEqual(ctx4.fixedrange(), [n1])

    def test_register_name(self):
        SF = StringField
        IF = IntegerField
        CF = ConstantField

        n1 = Number(1)
        n2 = Number(2)
        n3 = Number(3)
        n4 = Number(4)
        s1 = String("ab")
        s2 = String("cd")
        s3 = String("abcd")
        c1 = Function("ab", [])
        c2 = Function("cd", [])
        c3 = Function("abcd", [])
        # Test the register_name as a decorator
        cb1 = ContextBuilder()

        @cb1.register_name("addi")  # use function annotations
        def add1(a: IF, b: IF) -> IF:
            return a + b  # external signature

        self.assertEqual(add1(1, 2), 3)

        @cb1.register_name("adds", SF, SF, SF)
        def add2(a, b):
            return a + b

        self.assertEqual(add2("ab", "cd"), "abcd")

        # Non-decorator call - re-using a function but with a different signature
        cb1.register_name("addc", CF, CF, CF, add1)

        # Non-decorator call - setting a function with the function annotation
        cb1.register_name("addi_alt", add1)

        ctx1 = cb1.make_context()
        self.assertEqual(ctx1.addi(n1, n2), n3)
        self.assertEqual(ctx1.addi_alt(n1, n2), n3)
        self.assertEqual(ctx1.adds(s1, s2), s3)
        self.assertEqual(ctx1.addc(c1, c2), c3)

        # Things that should fail
        with self.assertRaises(TypeError) as ctx:
            self.assertEqual(ctx1.addc(s1, s2), s3)

        with self.assertRaises(TypeError) as ctx:
            self.assertEqual(ctx1.addc(s1, s2), c3)

        # Fails since add2 has no function annotations
        with self.assertRaises(TypeError) as ctx:
            cb1.register_name("addo", add2)

        # Function name already assigned
        with self.assertRaises(ValueError) as ctx:
            cb1.register_name("addi", add1)


# ------------------------------------------------------------------------------
# main
# ------------------------------------------------------------------------------
if __name__ == "__main__":
    raise RuntimeError("Cannot run modules")



# ------------------------------------------------------------------------------
# Unit tests for the clorm monkey patching
# ------------------------------------------------------------------------------
import unittest

import clingo as oclingo

import clorm.clingo as cclingo
from clorm import FactBase, IntegerField, Predicate, StringField, SymbolPredicateUnifier, ph1_
from clorm.clingo import Control, _expand_assumptions

from .support import add_program_string, check_errmsg

# ------------------------------------------------------------------------------
# ------------------------------------------------------------------------------

__all__ = ["ClingoTestCase"]

# ------------------------------------------------------------------------------
# ------------------------------------------------------------------------------


class ClingoTestCase(unittest.TestCase):
    def setUp(self):
        pass

    def tearDown(self):
        pass

    # --------------------------------------------------------------------------
    # Basic test of connecting to clingo without any wrapping
    # --------------------------------------------------------------------------
    def test_basic_clingo_connection(self):
        class F(Predicate):
            anum = IntegerField

        f1 = F(1)
        f2 = F(2)
        prgstr = "{}.\n{}.\n".format(f1, f2)
        ctrl = oclingo.Control()
        add_program_string(ctrl, prgstr)

        ctrl.ground([("base", [])])
        model = None
        with ctrl.solve(yield_=True) as sh:
            for m in sh:
                model = str(m)
        self.assertEqual(model, "{} {}".format(f1, f2))

    # --------------------------------------------------------------------------
    # Test processing clingo Model
    # --------------------------------------------------------------------------
    def test_control_model_integration(self):
        spu = SymbolPredicateUnifier()

        @spu.register
        class Afact(Predicate):
            num1 = IntegerField()
            num2 = IntegerField()
            str1 = StringField()

        @spu.register
        class Bfact(Predicate):
            num1 = IntegerField()
            str1 = StringField()

        af1 = Afact(1, 10, "bbb")
        af2 = Afact(2, 20, "aaa")
        af3 = Afact(3, 20, "aaa")
        bf1 = Bfact(1, "aaa")
        bf2 = Bfact(2, "bbb")

        fb1 = FactBase()
        fb1.add([af1, af2, af3, bf1, bf2])

        fb2 = None

        def on_model(model):
            nonlocal fb2
            self.assertTrue(model.contains(af1))
            self.assertTrue(model.contains(af1.raw))
            self.assertTrue(model.model_.contains(af1.raw))
            fb2 = model.facts(spu, atoms=True)

            # Check that the known attributes behave the same as the real model
            self.assertEqual(model.cost, model.model_.cost)
            self.assertEqual(model.number, model.model_.number)
            self.assertEqual(model.optimality_proven, model.model_.optimality_proven)
            self.assertEqual(model.thread_id, model.model_.thread_id)
            self.assertEqual(model.type, model.model_.type)

            # Note: the SolveControl object returned is created dynamically on
            # each call so will be different for both calls. So test that the
            # symbolic_atoms property is the same.
            sas1 = set(model.context.symbolic_atoms)
            sas2 = set(model.model_.context.symbolic_atoms)
            self.assertEqual(len(sas1), len(sas2))

            # Test that clorm.clingo.Model produces the correct string
            self.assertEqual(str(model), str(model.model_))

            if oclingo.__version__ < "5.5.0":
                self.assertEqual(repr(model), repr(model.model_))

        # Use the orignal clingo.Control object so that we can test the wrapper call
        ctrlX_ = oclingo.Control()
        ctrl = cclingo.Control(control_=ctrlX_)

        ctrl.add_facts(fb1)
        ctrl.ground([("base", [])])
        ctrl.solve(on_model=on_model)

        # Check that the known control attributes behave the same as the real control
        cfg1 = ctrl.configuration
        cfg2 = ctrl.control_.configuration

        self.assertEqual(len(cfg1), len(cfg2))
        self.assertEqual(set(cfg1.keys), set(cfg2.keys))
        sas1 = set(ctrl.symbolic_atoms)
        sas2 = set(ctrl.control_.symbolic_atoms)
        self.assertEqual(len(sas1), len(sas2))
        self.assertEqual(ctrl.is_conflicting, ctrl.control_.is_conflicting)
        stat1 = ctrl.statistics
        stat2 = ctrl.control_.statistics
        self.assertEqual(len(stat1), len(stat2))
        tas1 = ctrl.theory_atoms
        tas2 = ctrl.control_.theory_atoms
        self.assertEqual(len(list(tas1)), len(list(tas2)))

        # _control_add_facts works with both a list of facts (either
        # clorm.Predicate or clingo.Symbol instances) and a FactBase
        ctrl2 = Control()
        ctrl2.add_facts([af1, af2, af3.raw, bf1.raw, bf2])

        safact1 = fb2.query(Afact).where(Afact.num1 == ph1_)
        safact2 = fb2.query(Afact).where(Afact.num1 < ph1_)
        self.assertEqual(safact1.bind(1).singleton(), af1)
        self.assertEqual(safact1.bind(2).singleton(), af2)
        self.assertEqual(safact1.bind(3).singleton(), af3)
        self.assertEqual(set(list(safact2.bind(1).all())), set([]))
        self.assertEqual(set(list(safact2.bind(2).all())), set([af1]))
        self.assertEqual(set(list(safact2.bind(3).all())), set([af1, af2]))
        self.assertEqual(fb2.query(Bfact).where(Bfact.str1 == "aaa").singleton(), bf1)
        self.assertEqual(fb2.query(Bfact).where(Bfact.str1 == "bbb").singleton(), bf2)

    # --------------------------------------------------------------------------
    # Test the wrapping of control objects
    # --------------------------------------------------------------------------
    def test_control_and_model_wrapper(self):

        # Test a control wrapper of a wrapper
        ctrl0 = oclingo.Control()
        ctrl1 = cclingo.Control(control_=ctrl0)
        ctrl2 = cclingo.Control(control_=ctrl1)
        ctrl2.ground([("base", [])])

        with ctrl2.solve(yield_=True) as sh:
            tmp = [m.facts(unifier=[], atoms=True) for m in sh]
            self.assertEqual(len(tmp), 1)
            self.assertEqual(len(tmp[0]), 0)

        # Test a model wrapper
        ctrl = oclingo.Control()
        ctrl.ground([("base", [])])
        with ctrl.solve(yield_=True) as sh:
            for n, m_ in enumerate(sh):
                self.assertEqual(n, 0)
                m = cclingo.Model(model=m_, unifier=[])
                self.assertEqual(len(m.facts()), 0)

        # Test wrapping a bad object - missing attributes and functions

        # Missing ground function
        with self.assertRaises(AttributeError) as ctx:

            class Bad(object):
                def solve(self):
                    pass

            bad = Bad()
            ctrl = cclingo.Control(control_=bad)
        check_errmsg("'Bad' object has no attribute 'ground'", ctx)

        # Missing solve function
        with self.assertRaises(AttributeError) as ctx:

            class Bad(object):
                def ground(self):
                    pass

            bad = Bad()
            ctrl = cclingo.Control(control_=bad)
        check_errmsg("'Bad' object has no attribute 'solve'", ctx)

        # Ground is an attribute but not a function
        with self.assertRaises(AttributeError) as ctx:

            class Bad(object):
                def __init__(self):
                    self.ground = 4

                def solve(self):
                    pass

            bad = Bad()
            ctrl = cclingo.Control(control_=bad)
        check_errmsg(
            ("Wrapped object of type '{}' does not have a " "function 'ground()'").format(
                type(bad)
            ),
            ctx,
        )

        # Solve is an attribute but not a function
        with self.assertRaises(AttributeError) as ctx:

            class Bad(object):
                def __init__(self):
                    self.solve = 4

                def ground(self):
                    pass

            bad = Bad()
            ctrl = cclingo.Control(control_=bad)
        check_errmsg(
            ("Wrapped object of type '{}' does not have a " "function 'solve()'").format(
                type(bad)
            ),
            ctx,
        )

        # Model wrapper with no "symbols" function or attribute
        with self.assertRaises(AttributeError) as ctx:

            class Bad(object):
                def __init__(self):
                    self.symbols = 2

            bad = Bad()
            m = cclingo.Model(model=bad)
        check_errmsg(
            ("Wrapped object of type '{}' does not have a " "function 'symbols()'").format(
                type(bad)
            ),
            ctx,
        )

        with self.assertRaises(AttributeError) as ctx:

            class Bad(object):
                def __init__(self):
                    pass

            bad = Bad()
            m = cclingo.Model(model=bad)
        check_errmsg("'Bad' object has no attribute 'symbols'", ctx)

    def test_class_defs_are_equal(self):
        from clorm.clingo import (
            ClormControl,
            ClormModel,
            ClormSolveHandle,
            Control,
            Model,
            SolveHandle,
        )

        self.assertEqual(Control, ClormControl)
        self.assertEqual(Model, ClormModel)
        self.assertEqual(SolveHandle, ClormSolveHandle)


# ------------------------------------------------------------------------------
#
# ------------------------------------------------------------------------------


class Clingo54TestCase(unittest.TestCase):
    def setUp(self):
        pass

    def tearDown(self):
        pass

    # --------------------------------------------------------------------------
    # Test the wrapping of control objects
    # --------------------------------------------------------------------------
    def test_control_and_model_wrapper(self):

        # Test a control wrapper of a wrapper
        ctrl0 = oclingo.Control()
        ctrl1 = cclingo.Control(control_=ctrl0)
        ctrl2 = cclingo.Control(control_=ctrl1)
        ctrl2.ground([("base", [])])
        with ctrl2.solve(yield_=True) as sh:
            tmp = [m.facts(unifier=[], atoms=True) for m in sh]
            self.assertEqual(len(tmp), 1)
            self.assertEqual(len(tmp[0]), 0)

        # Test a model wrapper
        ctrl = oclingo.Control()
        ctrl.ground([("base", [])])
        with ctrl.solve(yield_=True) as sh:
            for n, m_ in enumerate(sh):
                self.assertEqual(n, 0)
                m = cclingo.Model(model=m_, unifier=[])
                self.assertEqual(len(m.facts()), 0)

        # Test wrapping a bad object - missing attributes and functions

        # Missing ground function
        with self.assertRaises(AttributeError) as ctx:

            class Bad(object):
                def solve(self):
                    pass

            bad = Bad()
            ctrl = cclingo.Control(control_=bad)
        check_errmsg("'Bad' object has no attribute 'ground'", ctx)

        # Missing solve function
        with self.assertRaises(AttributeError) as ctx:

            class Bad(object):
                def ground(self):
                    pass

            bad = Bad()
            ctrl = cclingo.Control(control_=bad)
        check_errmsg("'Bad' object has no attribute 'solve'", ctx)

        # Ground is an attribute but not a function
        with self.assertRaises(AttributeError) as ctx:

            class Bad(object):
                def __init__(self):
                    self.ground = 4

                def solve(self):
                    pass

            bad = Bad()
            ctrl = cclingo.Control(control_=bad)
        check_errmsg(
            ("Wrapped object of type '{}' does not have a " "function 'ground()'").format(
                type(bad)
            ),
            ctx,
        )

        # Solve is an attribute but not a function
        with self.assertRaises(AttributeError) as ctx:

            class Bad(object):
                def __init__(self):
                    self.solve = 4

                def ground(self):
                    pass

            bad = Bad()
            ctrl = cclingo.Control(control_=bad)
        check_errmsg(
            ("Wrapped object of type '{}' does not have a " "function 'solve()'").format(
                type(bad)
            ),
            ctx,
        )

        # Model wrapper with no "symbols" function or attribute
        with self.assertRaises(AttributeError) as ctx:

            class Bad(object):
                def __init__(self):
                    self.symbols = 2

            bad = Bad()
            m = cclingo.Model(model=bad)
        check_errmsg(
            ("Wrapped object of type '{}' does not have a " "function 'symbols()'").format(
                type(bad)
            ),
            ctx,
        )

        with self.assertRaises(AttributeError) as ctx:

            class Bad(object):
                def __init__(self):
                    pass

            bad = Bad()
            m = cclingo.Model(model=bad)
        check_errmsg("'Bad' object has no attribute 'symbols'", ctx)

    # --------------------------------------------------------------------------
    # Test processing clingo Model
    # --------------------------------------------------------------------------
    def test_control_model_integration(self):
        spu = SymbolPredicateUnifier()

        @spu.register
        class Afact(Predicate):
            num1 = IntegerField()
            num2 = IntegerField()
            str1 = StringField()

        @spu.register
        class Bfact(Predicate):
            num1 = IntegerField()
            str1 = StringField()

        af1 = Afact(1, 10, "bbb")
        af2 = Afact(2, 20, "aaa")
        af3 = Afact(3, 20, "aaa")
        bf1 = Bfact(1, "aaa")
        bf2 = Bfact(2, "bbb")

        fb1 = FactBase()
        fb1.add([af1, af2, af3, bf1, bf2])

        fb2 = None

        def on_model(model):
            nonlocal fb2
            self.assertTrue(model.contains(af1))
            self.assertTrue(model.contains(af1.raw))
            self.assertTrue(model.model_.contains(af1.raw))
            fb2 = model.facts(spu, atoms=True)

            # Check that the known attributes behave the same as the real model
            self.assertEqual(model.cost, model.model_.cost)
            self.assertEqual(model.number, model.model_.number)
            self.assertEqual(model.optimality_proven, model.model_.optimality_proven)
            self.assertEqual(model.thread_id, model.model_.thread_id)
            self.assertEqual(model.type, model.model_.type)

            # Note: the SolveControl object returned is created dynamically on
            # each call so will be different for both calls. So test that the
            # symbolic_atoms property is the same.
            sas1 = set(model.context.symbolic_atoms)
            sas2 = set(model.model_.context.symbolic_atoms)
            self.assertEqual(len(sas1), len(sas2))

            # Test that clorm.clingo.Model produces the correct string
            self.assertEqual(str(model), str(model.model_))
            if oclingo.__version__ < "5.5.0":
                self.assertEqual(repr(model), repr(model.model_))

        # Use the orignal clingo.Control object so that we can test the wrapper call
        ctrlX_ = oclingo.Control()
        ctrl = cclingo.Control(control_=ctrlX_)
        ctrl.add_facts(fb1)
        ctrl.ground([("base", [])])
        ctrl.solve(on_model=on_model)

        # Check that the known control attributes behave the same as the real control
        cfg1 = ctrl.configuration
        cfg2 = ctrl.control_.configuration
        self.assertEqual(len(cfg1), len(cfg2))
        self.assertEqual(set(cfg1.keys), set(cfg2.keys))
        sas1 = set(ctrl.symbolic_atoms)
        sas2 = set(ctrl.control_.symbolic_atoms)
        self.assertEqual(len(sas1), len(sas2))
        self.assertEqual(ctrl.is_conflicting, ctrl.control_.is_conflicting)
        stat1 = ctrl.statistics
        stat2 = ctrl.control_.statistics
        self.assertEqual(len(stat1), len(stat2))
        tas1 = ctrl.theory_atoms
        tas2 = ctrl.control_.theory_atoms
        self.assertEqual(len(list(tas1)), len(list(tas2)))

        # _control_add_facts works with both a list of facts (either
        # clorm.Predicate or clingo.Symbol instances) and a FactBase
        ctrl2 = Control()
        ctrl2.add_facts([af1, af2, af3.raw, bf1.raw, bf2])

        safact1 = fb2.query(Afact).where(Afact.num1 == ph1_)
        safact2 = fb2.query(Afact).where(Afact.num1 < ph1_)
        self.assertEqual(safact1.bind(1).singleton(), af1)
        self.assertEqual(safact1.bind(2).singleton(), af2)
        self.assertEqual(safact1.bind(3).singleton(), af3)
        self.assertEqual(set(list(safact2.bind(1).all())), set([]))
        self.assertEqual(set(list(safact2.bind(2).all())), set([af1]))
        self.assertEqual(set(list(safact2.bind(3).all())), set([af1, af2]))
        self.assertEqual(fb2.query(Bfact).where(Bfact.str1 == "aaa").singleton(), bf1)
        self.assertEqual(fb2.query(Bfact).where(Bfact.str1 == "bbb").singleton(), bf2)

        # Now test a select

    # --------------------------------------------------------------------------
    # Test processing clingo Model
    # --------------------------------------------------------------------------
    def test_clingo_to_clorm_model_integration(self):
        spu = SymbolPredicateUnifier()

        @spu.register
        class Afact(Predicate):
            num1 = IntegerField()

        af1 = Afact(1)
        af2 = Afact(2)

        fb1 = FactBase()
        fb1.add([af1, af2])

        def on_model(model):
            cmodel = cclingo.Model(model=model)
            self.assertTrue(cmodel.contains(af1))
            self.assertTrue(model.contains(af1.raw))

        # Use the orignal clingo.Control object so that we can test the model wrapper
        ctrl = cclingo.Control()
        ctrl.add_facts(fb1)
        octrl = ctrl.control_
        octrl.ground([("base", [])])
        octrl.solve(on_model=on_model)

    # --------------------------------------------------------------------------
    # Test passing a SymbolPredicateUnifier to the model constructors
    # --------------------------------------------------------------------------
    def test_model_default_spu(self):
        spu = SymbolPredicateUnifier()

        @spu.register
        class Afact(Predicate):
            num1 = IntegerField()

        af1 = Afact(1)
        af2 = Afact(2)

        # Test that the function works correctly when an spu is passed to the
        # Model constructor
        def on_model1(model):
            cmodel = cclingo.Model(model=model, unifier=spu)
            fb = cmodel.facts(atoms=True)
            self.assertEqual(len(fb.facts()), 2)

        ctrl = cclingo.Control()
        ctrl.add_facts([af1, af2])
        ctrl.ground([("base", [])])
        ctrl.control_.solve(on_model=on_model1)

        # Test that the function works correctly when a unifier list is passed
        # to the Model constructor
        def on_model1(model):
            cmodel = cclingo.Model(model=model, unifier=[Afact])
            fb = cmodel.facts(atoms=True)
            self.assertEqual(len(fb.facts()), 2)

        ctrl = cclingo.Control()
        ctrl.add_facts([af1, af2])
        ctrl.ground([("base", [])])
        ctrl.control_.solve(on_model=on_model1)

        # Test that an empty unifier list is valid (even if not useful)
        def on_model2(model):
            cmodel = cclingo.Model(model=model, unifier=[])
            fb = cmodel.facts(atoms=True)
            self.assertEqual(len(fb.facts()), 0)

        ctrl = cclingo.Control()
        ctrl.add_facts([af1, af2])
        ctrl.ground([("base", [])])
        ctrl.control_.solve(on_model=on_model2)

        # Test that it fails correctly when no unifier is passed
        def on_model3(model):
            cmodel = cclingo.Model(model=model)
            with self.assertRaises(ValueError) as ctx:
                fb = cmodel.facts(atoms=True)

        ctrl.control_.solve(on_model=on_model3)

    # --------------------------------------------------------------------------
    # Test passing a SymbolPredicateUnifier to the control constructors and using the
    # on_model callback for solving
    # --------------------------------------------------------------------------
    def test_control_on_model_default_spu(self):
        spu = SymbolPredicateUnifier()

        @spu.register
        class Afact(Predicate):
            num1 = IntegerField()

        af1 = Afact(1)
        af2 = Afact(2)

        # Test that the function works correctly when an spu is passed via the
        # clorm.clingo.Control constructor and using the on_model callback
        def on_model1(model):
            fb = model.facts(atoms=True)
            self.assertEqual(len(fb.facts()), 2)

        ctrl = cclingo.Control(unifier=spu)
        ctrl.add_facts([af1, af2])
        ctrl.ground([("base", [])])
        ctrl.solve(on_model=on_model1)

        # Test that the function works correctly when a unifier list is passed
        # via the clorm.clingo.Control constructor and using the on_model
        # callback
        def on_model1(model):
            fb = model.facts(atoms=True)
            self.assertEqual(len(fb.facts()), 2)

        ctrl = cclingo.Control(unifier=[Afact])
        ctrl.add_facts([af1, af2])
        ctrl.ground([("base", [])])
        ctrl.solve(on_model=on_model1)

        # Test that an empty unifier list is still valid (even if not useful)
        def on_model2(model):
            fb = model.facts(atoms=True)
            self.assertEqual(len(fb.facts()), 0)

        ctrl = cclingo.Control(unifier=[])
        ctrl.add_facts([af1, af2])
        ctrl.ground([("base", [])])
        ctrl.solve(on_model=on_model2)

        # Test that it fails correctly when no spu is passed
        def on_model3(model):
            with self.assertRaises(ValueError) as ctx:
                fb = model.facts(atoms=True)

        ctrl = cclingo.Control()
        ctrl.add_facts([af1, af2])
        ctrl.ground([("base", [])])
        ctrl.solve(on_model=on_model3)

        # Now set the unifier and (checking the getter and setter) then re-run
        # the solver to show that it works
        ctrl.unifier = [Afact]
        spu = ctrl.unifier
        self.assertTrue(type(spu), SymbolPredicateUnifier)
        self.assertEqual(spu.predicates, (Afact,))
        self.assertEqual(spu.indexes, ())
        ctrl.solve(on_model=on_model1)

    # --------------------------------------------------------------------------
    # Test passing a SymbolPredicateUnifier to the control constructors and using a
    # solvehandle for solving
    # --------------------------------------------------------------------------
    def test_control_solvehandle_default_spu(self):
        spu = SymbolPredicateUnifier()

        @spu.register
        class Afact(Predicate):
            num1 = IntegerField()

        af1 = Afact(1)
        af2 = Afact(2)

        # Test that the function works correctly when an spu is passed via the
        # clorm.clingo.Control constructor and using the solvehandle
        def on_model1(model):
            fb = model.facts(atoms=True)
            self.assertEqual(len(fb.facts()), 2)

        ctrl = cclingo.Control(unifier=spu)
        ctrl.add_facts([af1, af2])
        ctrl.ground([("base", [])])
        with ctrl.solve(yield_=True) as sh:
            for m in sh:
                fb = m.facts(atoms=True)
                self.assertEqual(len(fb.facts()), 2)

        ctrl = cclingo.Control()
        ctrl.add_facts([af1, af2])
        ctrl.ground([("base", [])])
        with ctrl.solve(yield_=True) as sh:
            for m in sh:
                with self.assertRaises(ValueError) as ctx:
                    fb = m.facts(atoms=True)

    # --------------------------------------------------------------------------
    # Test the different argument combinations for clingo Model.facts()
    # --------------------------------------------------------------------------
    def test_model_facts_arguments(self):
        class Af(Predicate):
            num1 = IntegerField()

        af1 = Af(1)
        af2 = Af(2)

        ctrl = cclingo.Control()
        ctrl.add_facts([af1, af2])
        ctrl.ground([("base", [])])
        with ctrl.solve(yield_=True) as sh:
            m = next(iter(sh))
            fb = m.facts(unifier=[Af], atoms=True, raise_on_empty=True)
            self.assertEqual(len(fb.facts()), 2)
            fb = m.facts([Af], True, True, True, raise_on_empty=True)
            self.assertEqual(len(fb.facts()), 2)

            with self.assertRaises(TypeError) as ctx:
                fb = m.facts([Af], unifier=[Af])
            check_errmsg("facts() got multiple values for argument 'unifier'", ctx)

            with self.assertRaises(TypeError) as ctx:
                fb = m.facts([Af], True, True, True, True, True, True, raise_on_empty=True)
            check_errmsg("facts() got multiple values for argument 'raise_on_empty'", ctx)

    # --------------------------------------------------------------------------
    # Test processing clingo Model
    # --------------------------------------------------------------------------
    def test_model_facts(self):
        spu = SymbolPredicateUnifier()

        @spu.register
        class Afact(Predicate):
            num1 = IntegerField()
            num2 = IntegerField()
            str1 = StringField()

        class Bfact(Predicate):
            num1 = IntegerField()
            str1 = StringField()

        af1 = Afact(1, 10, "bbb")
        af2 = Afact(2, 20, "aaa")
        af3 = Afact(3, 20, "aaa")
        bf1 = Bfact(1, "aaa")
        bf2 = Bfact(2, "bbb")

        def on_model1(model):
            fb = model.facts(spu, atoms=True, raise_on_empty=True)
            self.assertEqual(len(fb.facts()), 3)  # spu only imports Afact

        ctrl = cclingo.Control()
        ctrl.add_facts([af1, af2, af3, bf1, bf2])
        ctrl.ground([("base", [])])
        ctrl.solve(on_model=on_model1)

        spu2 = SymbolPredicateUnifier()

        def on_model2(model):
            # Note: because of the delayed initialisation you have to do
            # something with the factbase to get it to raise the error.
            with self.assertRaises(ValueError) as ctx:
                fb = model.facts(spu2, atoms=True, raise_on_empty=True)
                self.assertEqual(len(fb.facts()), 0)

        ctrl = cclingo.Control()
        ctrl.add_facts([bf1, bf2])
        ctrl.ground([("base", [])])
        ctrl.solve(on_model=on_model2)

    # --------------------------------------------------------------------------
    # Test the solvehandle
    # --------------------------------------------------------------------------
    def test_solvehandle_wrapper(self):
        spu = SymbolPredicateUnifier()

        @spu.register
        class Fact(Predicate):
            num1 = IntegerField()

            class Meta:
                name = "f"

        prgstr = """{ g(N) : f(N) } = 1."""
        f1 = Fact(1)
        f2 = Fact(2)
        f3 = Fact(3)
        ctrl = cclingo.Control(["-n 0"])
        add_program_string(ctrl, prgstr)
        ctrl.add_facts([f1, f2, f3])
        ctrl.ground([("base", [])])

        with ctrl.solve(yield_=True) as sh:
            self.assertTrue(isinstance(sh, cclingo.SolveHandle))
            self.assertFalse(isinstance(sh, oclingo.SolveHandle))
            self.assertTrue(sh.solvehandle_)
            num_models = 0
            for m in sh:
                self.assertTrue(isinstance(m, cclingo.Model))
                self.assertFalse(isinstance(m, oclingo.Model))
                num_models += 1
            self.assertEqual(num_models, 3)

    # ----------------------------------------------------------------------------------------
    # Test expanding the assumptions - note: the order matters so for an input list of
    # predicates or symbols the output list is the corresponding symbols in the same order.
    # ----------------------------------------------------------------------------------------
    def test_expand_assumptions(self):
        class F(Predicate):
            num1 = IntegerField()

        class G(Predicate):
            num1 = IntegerField()

        f1 = F(1)
        f2 = F(2)
        g1 = G(1)

        r = _expand_assumptions([(f1, True), (g1, False)])
        self.assertEqual(r, [(f1.raw, True), (g1.raw, False)])

        r = _expand_assumptions([(FactBase([f1, f2]), True), (set([g1]), False)])
        self.assertEqual(r, [(f1.raw, True), (f2.raw, True), (g1.raw, False)])

        with self.assertRaises(TypeError) as ctx:
            _expand_assumptions([g1])
        with self.assertRaises(TypeError) as ctx:
            _expand_assumptions(g1)

    # ----------------------------------------------------------------------------------------
    # Test the solvehandle
    # ----------------------------------------------------------------------------------------
    def test_solve_with_assumptions_simple(self):
        spu = SymbolPredicateUnifier()

        @spu.register
        class F(Predicate):
            num1 = IntegerField()

        @spu.register
        class G(Predicate):
            num1 = IntegerField()

        prgstr = """{ g(N) : f(N) } = 1."""
        f1 = F(1)
        f2 = F(2)
        f3 = F(3)
        g1 = G(1)
        g2 = G(2)
        g3 = G(3)
        ctrl = cclingo.Control(["-n 0"])
        add_program_string(ctrl, prgstr)
        ctrl.add_facts([f1, f2, f3])
        ctrl.ground([("base", [])])

        num_models = 0

        def on_modelT(m):
            nonlocal num_models
            fb = m.facts(spu, atoms=True)
            self.assertTrue(g1 in fb)
            num_models += 1

        def on_modelF(m):
            nonlocal num_models
            fb = m.facts(spu, atoms=True)
            self.assertTrue(g1 not in fb)
            self.assertFalse(g1 in fb)
            num_models += 1

        num_models = 0
        ctrl.solve(on_model=on_modelT, assumptions=[(g1, True)])
        self.assertEqual(num_models, 1)
        num_models = 0
        ctrl.solve(on_model=on_modelT, assumptions=[(g1.raw, True)])
        self.assertEqual(num_models, 1)
        num_models = 0
        ctrl.solve(on_model=on_modelF, assumptions=[(g1, False)])
        self.assertEqual(num_models, 2)
        fb2 = FactBase([g1])
        num_models = 0
        ctrl.solve(on_model=on_modelT, assumptions=[(fb2, True)])
        self.assertEqual(num_models, 1)

    # --------------------------------------------------------------------------
    # Test the solvehandle
    # --------------------------------------------------------------------------
    def test_solve_with_assumptions_complex(self):
        class F(Predicate):
            num1 = IntegerField()

        class G(Predicate):
            num1 = IntegerField()

        prgstr = """ 1 { g(N) : f(N) } 2."""
        f1 = F(1)
        f2 = F(2)
        f3 = F(3)
        g1 = G(1)
        g2 = G(2)
        g3 = G(3)
        ctrl = cclingo.Control(["-n 0"], unifier=[G])
        add_program_string(ctrl, prgstr)
        ctrl.add_facts([f1, f2, f3])
        ctrl.ground([("base", [])])

        num_models = 0

        def on_model(m):
            nonlocal num_models
            fb = m.facts(atoms=True)
            self.assertTrue(len(fb) <= 2)
            self.assertTrue(len(fb) >= 1)
            num_models += 1

        num_models = 0
        if oclingo.__version__ >= "5.5.0":
            ctrl.solve(on_model=on_model, assumptions=[])
        else:
            ctrl.solve(on_model=on_model, assumptions=None)
        self.assertEqual(num_models, 6)
        num_models = 0
        ctrl.solve(on_model=on_model)
        self.assertEqual(num_models, 6)

        num_models = 0
        ctrl.solve(on_model=on_model, assumptions=[(g1, True)])
        self.assertEqual(num_models, 3)

        # Mixing raw symbol and predicate in a set
        num_models = 0
        ctrl.solve(on_model=on_model, assumptions=[(set([g1.raw, g2]), True)])
        self.assertEqual(num_models, 1)
        num_models = 0
        ctrl.solve(on_model=on_model, assumptions=[(FactBase([g1, g2]), True)])
        self.assertEqual(num_models, 1)
        num_models = 0
        ctrl.solve(on_model=on_model, assumptions=[(FactBase([g1]), True), (set([g2]), False)])
        self.assertEqual(num_models, 2)

        num_models = 0
        ctrl.solve(on_model=on_model, assumptions=[(FactBase([g1]), True), (set([g2]), False)])
        self.assertEqual(num_models, 2)

    # --------------------------------------------------------------------------
    # Test the solve
    # --------------------------------------------------------------------------
    def test_solve_with_on_finish(self):
        spu = SymbolPredicateUnifier()

        @spu.register
        class F(Predicate):
            num1 = IntegerField()

        f1 = F(1)
        f2 = F(2)
        f3 = F(3)
        infb = FactBase([f1, f2, f2])
        ctrl = cclingo.Control(["-n 0"])
        ctrl.add_facts(infb)
        ctrl.ground([("base", [])])

        called = False

        def on_model(m):
            nonlocal called
            outfb = m.facts(spu, atoms=True)
            self.assertEqual(infb, outfb)
            self.assertFalse(called)
            called = True

        def on_finish(sr):
            self.assertTrue(sr.satisfiable)
            self.assertFalse(sr.unsatisfiable)

        sr = ctrl.solve(on_model=on_model, on_finish=on_finish)
        self.assertTrue(sr.satisfiable)
        self.assertFalse(sr.unsatisfiable)

    # --------------------------------------------------------------------------
    # Test the solve
    # --------------------------------------------------------------------------
    def test_solve_with_on_statistics(self):
        class F(Predicate):
            num1 = IntegerField()

        f1 = F(1)
        f2 = F(2)
        f3 = F(3)
        infb = FactBase([f1, f2, f2])
        ctrl = cclingo.Control(["-n 0"])
        ctrl.add_facts(infb)
        ctrl.ground([("base", [])])

        def on_model(model):
            nonlocal om_called
            om_called = True

        def on_statistics(stp, acc):
            nonlocal os_called
            self.assertEqual(type(stp), oclingo.StatisticsMap)
            os_called = True

        # Calling using positional arguments
        om_called = False
        os_called = False
        if oclingo.__version__ < "5.5.0":
            sr = ctrl.solve([], on_model, on_statistics)
        else:
            sr = ctrl.solve([], on_model, None, on_statistics)
        self.assertTrue(om_called)
        self.assertTrue(os_called)

        # Calling using keyword arguments
        om_called = False
        os_called = False
        sr = ctrl.solve(on_statistics=on_statistics)
        self.assertFalse(om_called)
        self.assertTrue(os_called)

    # --------------------------------------------------------------------------
    # Test the solve with on_unsat
    # --------------------------------------------------------------------------
    def test_solve_with_on_unsat(self):
        # There is no on_unsat or on_core functions for older versions.
        if oclingo.__version__ < "5.5.0":
            return

        # Program to test optimisation for on_unsat callback - note: setting the appropriate
        # optimisation strategy is important to trigger the callback.
        prgstr = """
1 { p(X); q(X) } 1 :- X=1..3.
#minimize { 1,p,X: p(X); 1,q,X: q(X) }.
        """

        ctrl = cclingo.Control()
        conf = ctrl.configuration
        conf.solver.opt_strategy = "usc,oll,0"
        add_program_string(ctrl, prgstr)
        ctrl.ground([("base", [])])

        def on_unsat(ints):
            nonlocal ou_called
            ou_called = True

        # Calling using positional arguments
        ou_called = False
        sr = ctrl.solve([], None, on_unsat)
        self.assertTrue(ou_called)

        # Caling using keyword arguments
        ou_called = False
        sr = ctrl.solve(on_unsat=on_unsat)
        self.assertTrue(ou_called)

    # --------------------------------------------------------------------------
    # Test the solve with on_core
    # --------------------------------------------------------------------------
    def test_solve_with_on_core(self):
        # There is no on_unsat or on_core functions for older versions.
        if oclingo.__version__ < "5.5.0":
            return

        class F(Predicate):
            num1 = IntegerField()

        # Program to test unsatisfiable cores
        prgstr = """
2 {f(1..3)} 2.
"""
        all_facts = FactBase([F(1), F(2), F(3)])
        ctrl = cclingo.Control(unifier=[F])
        add_program_string(ctrl, prgstr)
        ctrl.ground([("base", [])])
        all_literals = [ctrl.symbolic_atoms[f.raw].literal for f in all_facts]

        def on_core(ints):
            nonlocal oc_called
            assert set(all_literals) == set(ints)
            oc_called = True

        # Calling using positional arguments
        oc_called = False  #
        sr = ctrl.solve([(all_facts, True)], None, None, None, None, on_core)
        self.assertTrue(oc_called)

        # Calling using keyword arguments
        oc_called = False
        sr = ctrl.solve(assumptions=[(all_facts, True)], on_core=on_core)
        self.assertTrue(oc_called)

    # --------------------------------------------------------------------------
    # Test accessing some other control variables that are not explcitly wrapped
    # --------------------------------------------------------------------------
    def test_control_access_others(self):
        class F(Predicate):
            num1 = IntegerField()

        f1 = F(1)
        f2 = F(2)
        f3 = F(3)
        infb = FactBase([f1, f2])
        ctrl = cclingo.Control(["-n 0"])
        ctrl.add_facts(infb)
        ctrl.ground([("base", [])])
        self.assertTrue(ctrl.symbolic_atoms[f1.raw])

    # --------------------------------------------------------------------------
    # Test the solvehandle
    # --------------------------------------------------------------------------
    def test_solve_returning_solvehandle(self):
        spu = SymbolPredicateUnifier()

        @spu.register
        class F(Predicate):
            num1 = IntegerField()

        f1 = F(1)
        f2 = F(2)
        f3 = F(3)
        infb = FactBase([f1, f2, f3])
        ctrl = cclingo.Control(["-n 0"])
        ctrl.add_facts(infb)
        ctrl.ground([("base", [])])

        done = False

        def on_model(m):
            nonlocal done
            outfb = m.facts(spu, atoms=True)
            self.assertEqual(infb, outfb)
            self.assertFalse(done)
            done = True

        if oclingo.__version__ > "5.3.1":
            asynckw = "async_"
        else:
            asynckw = "async"

        # Test the async mode
        kwargs = {"on_model": on_model, asynckw: True}
        done = False
        sh = ctrl.solve(**kwargs)
        self.assertTrue(isinstance(sh, cclingo.SolveHandle))
        sh.get()
        self.assertTrue(done)

        # Test the yield mode
        kwargs = {"on_model": on_model, "yield_": True}
        done = False
        sh = ctrl.solve(**kwargs)
        self.assertTrue(isinstance(sh, cclingo.SolveHandle))
        count = 0
        for m in sh:
            count += 1
            outfb = m.facts(spu, atoms=True)
            self.assertEqual(infb, outfb)
        self.assertEqual(count, 1)
        self.assertTrue(done)

        # Test both async and yield mode
        kwargs = {"on_model": on_model, asynckw: True, "yield_": True}
        done = False
        sh = ctrl.solve(**kwargs)
        self.assertTrue(isinstance(sh, cclingo.SolveHandle))
        count = 0
        for m in sh:
            count += 1
            outfb = m.facts(spu, atoms=True)
            self.assertEqual(infb, outfb)
        self.assertEqual(count, 1)
        self.assertTrue(done)

    # --------------------------------------------------------------------------
    # Test bad arguments
    # --------------------------------------------------------------------------
    def test_bad_solve_arguments(self):
        spu = SymbolPredicateUnifier()

        @spu.register
        class Fact(Predicate):
            num1 = IntegerField()

            class Meta:
                name = "f"

        f1 = Fact(1)
        f2 = Fact(2)
        f3 = Fact(3)
        ctrl = cclingo.Control(["-n 0"])
        ctrl.add_facts([f1, f2, f3])
        ctrl.ground([("base", [])])

        with self.assertRaises(TypeError) as ctx:
            ctrl.solve(assump=[f1])

        with self.assertRaises(TypeError) as ctx:
            ctrl.solve([f1], assumptions=[f1])

        with self.assertRaises(TypeError) as ctx:
            ctrl.solve([f1])

    # --------------------------------------------------------------------------
    # Test assign external
    # --------------------------------------------------------------------------
    def test_assign_and_release_external(self):
        class F(Predicate):
            num1 = IntegerField()

        class G(Predicate):
            num1 = IntegerField()

        prgstr = """
#external f(1..3).
g(N) :- f(N)."""

        f1 = F(1)
        f2 = F(2)
        f3 = F(3)
        g1 = G(1)
        g2 = G(2)
        g3 = G(3)
        ctrl = cclingo.Control(unifier=[F, G])
        add_program_string(ctrl, prgstr)
        ctrl.ground([("base", [])])

        # Assign external for a factbase
        ctrl.assign_external(FactBase([f1, f2, f3]), True)
        with ctrl.solve(yield_=True) as sh:
            m = list(sh)[0]
            fb = m.facts(atoms=True)
            self.assertEqual(fb, FactBase([f1, f2, f3, g1, g2, g3]))

        # Assign external for a single clorm fact
        ctrl.assign_external(f1, False)
        with ctrl.solve(yield_=True) as sh:
            m = list(sh)[0]
            fb = m.facts(atoms=True)
            self.assertEqual(fb, FactBase([f2, f3, g2, g3]))

        # Assign external for a single clingo symbol
        ctrl.assign_external(f2.raw, False)
        with ctrl.solve(yield_=True) as sh:
            m = list(sh)[0]
            fb = m.facts(atoms=True)
            self.assertEqual(fb, FactBase([f3, g3]))

        # Back to all true so we can test release_external
        # Assign external for a factbase
        ctrl.assign_external(FactBase([f1, f2, f3]), True)
        with ctrl.solve(yield_=True) as sh:
            m = list(sh)[0]
            fb = m.facts(atoms=True)
            self.assertEqual(fb, FactBase([f1, f2, f3, g1, g2, g3]))

        # Release external for a FactBase
        ctrl.release_external(FactBase([f1]))
        with ctrl.solve(yield_=True) as sh:
            m = list(sh)[0]
            fb = m.facts(atoms=True)
            self.assertEqual(fb, FactBase([f2, f3, g2, g3]))

        # Release external for a single clorm fact
        ctrl.release_external(f2)
        with ctrl.solve(yield_=True) as sh:
            m = list(sh)[0]
            fb = m.facts(atoms=True)
            self.assertEqual(fb, FactBase([f3, g3]))

        # Release external for a single clingo symbol
        ctrl.release_external(f3.raw)
        with ctrl.solve(yield_=True) as sh:
            m = list(sh)[0]
            fb = m.facts(atoms=True)
            self.assertEqual(fb, FactBase())


# ------------------------------------------------------------------------------
#
# ------------------------------------------------------------------------------
# ------------------------------------------------------------------------------
# main
# ------------------------------------------------------------------------------
if __name__ == "__main__":
    raise RuntimeError("Cannot run modules")
